/*
**	ai.c
*/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

#define BS 8192
#define OBS 8192
#define AMP 32000

char snd_dev_lbl[32][32]= SOUND_DEVICE_LABELS;
short obf[BS];
short ibf[OBS];
int channels;
int audio_fd = -1;
int mixer_fd = -1;
int play_precision;
int play_stereo;
int play_sample_rate;
int be= 0;
int freq=44100;
int volume=100;
int noc=2;
FILE *os=NULL;
FILE *is=NULL;
/*
int mode= O_RDWR;
int mode= O_WRONLY;
*/
int mode= O_RDONLY;

void exit1(char*p) { fprintf(stderr,p); exit(1); }
void usage(char *p) {
	fprintf(stderr,"Usage:%s [-][-b][-fN][-vN] file \n",p);
	fprintf(stderr,"-   from stdin to stdout\n");
	fprintf(stderr,"-m  mono\n");
	fprintf(stderr,"-b  Big Endian\n");
	fprintf(stderr,"-fN N: sampling frequency [Hz]\n");
	fprintf(stderr,"-vN N: [1,100] volume\n");
	exit(1);
}


main(int argc,char **argv)
{
	char *prog= argv[0];
	argc--; argv++;
	while (argc) {
		if (strncmp(*argv,"-v",2)==0) volume= atoi(*argv+2);
		else if (strncmp(*argv, "-f",2)==0) freq= atoi(*argv+2);
		else if (strcmp(*argv, "-b")==0) be= 1;
		else if (strcmp(*argv, "-m")==0) noc= 1;
		else if (strcmp(*argv, "-") ==0) { os= stdout; is= stdin; }
		else if (**argv == '-') {
			fprintf(stderr, "unknown option: %s\n", *argv);
			usage(prog);
		}
	else break;
	argc--; argv++;
	}
	if(argc--==1){
       	//if ((os = fopen(*argv++,"r")) == NULL) exit1("can't read open\n");
       	if ((is = fopen(*argv++,"w")) == NULL)  exit1("can't write open\n");
	}
	else if(os!=stdout) usage(prog);
	soundin();
}

soundin()
{
	int l;
	open_audio (noc,freq,volume);
	while(l= read(audio_fd,ibf,sizeof(ibf))) fwrite (ibf,1,l,is);
	close(audio_fd);
	close(mixer_fd);
}

soundout()
{
	char *buf;
	int l, len, n;
	open_audio (noc,freq,volume);
	while(l= fread(obf,1,sizeof(obf),os)){
		if(be)big_endian();
		buf = (char *)obf; len = l;
		while(len) {
			if ((n= write (audio_fd, buf, len)) < 0) {
				if (errno != EINTR)
				{perror("write to /dev/dsp failed");exit(1);}
				n= 0;
			}
		len-= n; buf+= n;
		}
	}
	close(audio_fd);
	close(mixer_fd);
}

soundio() {
	int l=0;
	int r=0;
	int of= BS-16;
	int a= AMP;

	obf[of  ]=  a;	obf[of+1]=  a;
	obf[of+2]= -a;	obf[of+3]= -a;
	obf[of+4]=  a;	obf[of+5]=  a;
	obf[of+6]= -a;	obf[of+7]= -a;

	open_audio (noc,freq,volume);

	while(l<sizeof(obf)) l+= write (audio_fd, obf+l, sizeof(obf)-l);

	close(audio_fd);
	close(mixer_fd);

	open_audio (noc,freq,volume);

	while(r<sizeof(ibf)) r+= read(audio_fd,ibf+r,sizeof(ibf)-r);

	fwrite (ibf,1,sizeof(ibf),is);

	close(audio_fd);
	close(mixer_fd);
}


big_endian() {
	int i;
	for(i=0;i<BS;i++){
	unsigned int d= (unsigned short)obf[i];
	obf[i]= d<<8 | d>>8;
	}
}


open_audio (noc,freq,volume)
int noc;	// nomber_of_channels
int freq;	//sampling frequency [kHz]
int volume;	// volume 0-100
{
	int i, fd, flags, ret, devices, vol_chan;
	if ((fd = open ("/dev/dsp", mode | O_NDELAY, 0)) < 0)
		exit1("/dev/dsp is busy or can't be opened\n");
	audio_fd= fd;
	// turn NDELAY mode off:
	if ((flags = fcntl (fd, F_GETFL, 0)) < 0)
		exit1("fcntl F_GETFL on /dev/audio failed\n");
	flags &= ~O_NDELAY;
	if (fcntl (fd, F_SETFL, flags) < 0)
		exit1("fcntl F_SETFL on /dev/audio failed\n");
	channels = noc;
	play_precision= 16;
	play_stereo = channels-1;
	play_sample_rate = freq;
	if(	ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &play_precision) == -1 ||
		ioctl(fd, SNDCTL_DSP_STEREO, &play_stereo) == -1 ||
		ioctl(fd, SNDCTL_DSP_SPEED, &play_sample_rate) == -1
	)	exit1("configuration of /dev/dsp failed\n");

	if ((mixer_fd= fd = open ("/dev/mixer", O_WRONLY | O_NDELAY, 0)) < 0)
		exit1("/dev/mixer is busy or can't be opened\n");

	ret = ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devices);
	if (ret == -1) devices = 0;

	for(i=0;i<17;i++){
		if(devices & (1<<i)) {
			vol_chan= devices & (1<<i);
			fprintf(stderr,"%d= %s",i,snd_dev_lbl[i]);
			if( ioctl(fd, MIXER_READ(vol_chan), &volume) == -1)
				fprintf(stderr,"\n");
			else
			fprintf(stderr,"vol=%d and %d\n", volume>>8, volume&255);
		}
	}
	if (devices & (1 << SOUND_MIXER_PCM))vol_chan = SOUND_MIXER_PCM;
	else				     vol_chan = SOUND_MIXER_VOLUME;
	fprintf(stderr,"vol_chan= %d\n",vol_chan);
	volume |= volume << 8;
	if( ioctl(fd, MIXER_WRITE(vol_chan), &volume) == -1)
		exit1("configuration of /dev/mixer failed\n");
	if (devices & (1 << SOUND_MIXER_LINE))	vol_chan = SOUND_MIXER_LINE;
	else					vol_chan = SOUND_MIXER_VOLUME;

	fprintf(stderr,"vol_chan= %d\n",vol_chan);

	if( ioctl(fd, MIXER_WRITE(vol_chan), &volume) == -1)
		exit1("configuration of /dev/mixer failed\n");

	for(i=0;i<17;i++){
		if(devices & (1<<i)) {
			vol_chan= devices & (1<<i);
			fprintf(stderr,"%d= %s",i,snd_dev_lbl[i]);
			if( ioctl(fd, MIXER_READ(vol_chan), &volume) == -1) fprintf(stderr,"\n");
			else fprintf(stderr,"vol=%d and %d\n", volume>>8, volume&255);
		}
	}
}

