/*
**	audio.c
*/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

//#define BS (2*1152)
#define BS 8192

short buffer[BS];
int channels;
int audio_fd = -1;
int mixer_fd = -1;
int play_precision;
int play_stereo;
int play_sample_rate;
int be= 0;
int freq= 44100;
int volume= 100;
int noc=2;

void
usage(char *p){
	fprintf(stderr,"Usage:%s [-][-b][-fN][-vN] [file] \n",p);
	fprintf(stderr,"-   from stdin\n");
	fprintf(stderr,"-m  mono\n");
	fprintf(stderr,"-b  Big Endian\n");
	fprintf(stderr,"-fN N: sampling frequency [Hz]\n");
	fprintf(stderr,"-vN N: [1,100] volume\n");
	exit(1);
}

void
exit1(char*p) { fprintf(stderr,p); exit(1); }
void
exit2(char*p,char*q) { fprintf(stderr,p,q); exit(1); }

int
big_endian()
{
	int i;
	unsigned int d;
	for(i=0;i<BS;i++){
		d= (unsigned short)buffer[i];
		buffer[i]= d<<8 | d>>8;
	}
}

int
open_audio_device (noc,freq,volume)
int noc;	// nomber_of_channels
int freq;	// sampling frequency [kHz]
int volume;	// volume 0-100
{
	int i, fd, flags, ret, devices, vol_chan;

	if ((fd = open ("/dev/dsp", O_WRONLY | O_NDELAY, 0))< 0)
		exit1("audio device /dev/dsp is busy or can't be opened\n");
	// turn NDELAY mode off: 
	if ((flags = fcntl (fd, F_GETFL, 0))< 0) exit1("fcntl F_GETFL on /dev/audio failed\n");
	flags &= ~O_NDELAY;
	if (fcntl (fd, F_SETFL, flags)< 0) exit1("fcntl F_SETFL on /dev/audio failed\n");
	audio_fd= fd;
	channels = noc;
	// configure the device:
	play_precision= 16;
	play_stereo = channels-1;
	play_sample_rate = freq;
	if(
		ioctl(audio_fd, SNDCTL_DSP_SAMPLESIZE, &play_precision) == -1 ||
		ioctl(audio_fd, SNDCTL_DSP_STEREO, &play_stereo) == -1 ||
		ioctl(audio_fd, SNDCTL_DSP_SPEED, &play_sample_rate) == -1
	) exit1("configuration of /dev/dsp failed\n");

	if ((fd = open("/dev/mixer", O_WRONLY | O_NDELAY, 0))< 0)
		exit1("audio device /dev/mixer is busy or can't be opened\n");

	mixer_fd= fd;
	ret = ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &devices);
	if (ret == -1) devices = 0;
	if (devices & (1 << SOUND_MIXER_PCM))
		vol_chan = SOUND_MIXER_PCM;
	else        vol_chan = SOUND_MIXER_VOLUME;
	volume |= volume << 8;
	if( ioctl(mixer_fd, MIXER_WRITE(vol_chan), &volume) == -1)
	exit1("configuration of /dev/mixer failed\n");
}

int
main(argc, argv)
int     argc;
char    *argv[];
{
	char *prog;
	char *bsname= NULL;
	FILE *bs= NULL;
	int l;

	prog= argv[0];
	argc--; argv++;
        while (argc) {
		if (strncmp(*argv,"-v",2)==0) volume= atoi(*argv+2);
		else if (strncmp(*argv, "-f",2)==0) freq= atoi(*argv+2);
		else if (strcmp(*argv, "-b")==0) be= 1;
		else if (strcmp(*argv, "-m")==0) noc= 1;
		else if (strcmp(*argv, "-") ==0) bs= stdin;
		else if (**argv == '-') {
			fprintf(stderr, "unknown option: %s\n", *argv);
			usage(prog);
		}
		else break;
		argc--; argv++;
	}
	if(argc--){
		bsname= *argv++;
		if ((bs = fopen(bsname,"r")) == NULL)
			exit2("can't open %s\n",bsname);
	} else {
		if(bs!=stdin)usage(prog);
	}
	open_audio_device (noc,freq,volume);
	while(l= fread(buffer,1,sizeof(buffer),bs)){
	//while((l= read(fileno(bs),buffer,sizeof(buffer))) > 0){
		char *buf;
		int len, n;

		if(be)big_endian();
		buf = (char *)buffer; len = l;
		while(len) {
			if ((n= write (audio_fd, buf, len)) < 0) {
				if (errno != EINTR)
				{perror("write to /dev/dsp failed");exit(1);}
				n= 0;
			}
			len-= n; buf+= n;
		}
	}
	close(audio_fd);
	close(mixer_fd);
}
