/*
**	aout.c
*/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <math.h>

#define MAXL 1024	/* max impulse response	length		*/
#define BS (2*8192)	/* music buffer size noc*8192		*/
#define IS (2*MAXL)	/* Impulse Size:1 channel MAXL sample	*/
#define MX ((1<<15)-1)	/* max short value			*/
#define MN (-(1<<15))	/* min short value			*/
#define IR "/tmp/ir"	/* impulse response file		*/
#define AC "/tmp/ac"	/* audio control file			*/

int	buf[IS+BS];	/* preceding IS is for last output	*/
double	energy;		/* energy 				*/
int	asum; 		/* absolute sum of impulse		*/
short	spl[MAXL];	/* impulse responce data (short)	*/
int	ipl[MAXL];	/* impulse responce data (int)		*/
short	ibf[4*BS];	/* input buffer				*/
short	obf[BS];	/* FIR output buffer			*/
int	ib= 0; 		/* ibf bank selecter 			*/
FILE	*bs= NULL;	/* music file pointer			*/
int	bz= BS;		/* initial buffer size			*/
short	*tail=NULL;
int	coeff= 64;	/* echo coeff 				*/

int alg= 0;
int be=  0;
int ph=  0;
int rv=  0;
int rvl= 1;
int irlen= 0;
int freq=44100;
int volume=100;
int noc= 2;
int gain= 16;
int echo= 0;
int echolen= 0;
int diff= 0;
int expd= 0;
int ffir=  0;
short fa[10]= {64,0,0,0,0,0,0,0,0,0};

void
usage(char *p){
	fprintf(stderr,"Usage:%s [-][-b][-fN][-vN] [file]\n",p);
	fprintf(stderr,"- from stdin, -m mono, -b: big endian\n");
	fprintf(stderr,"-a1: FIR, -a2: IIR, -a3: LPC\n");
	fprintf(stderr,"-a4: REV, -a5: DifRev -a6: Dif3Rev -a7: Trans \n");
	fprintf(stderr,"-eN:echo N sample\n");
	fprintf(stderr,"-fN N: sampling frequency [Hz]\n");
	fprintf(stderr,"-vN N: volume[1,100]\n");
	fprintf(stderr,"-pN Phaser: -16(mono) -0(def)- 16\n");
	exit(1);
}

main(argc,argv)
char **argv;
	{
	char *prg; FILE *fp; int i,l;

	prg= argv[0];
	//unlink(prg);
	argc--; argv++;
        while (argc) {
		     if (strncmp(*argv,"-v",2)==0) volume= atoi(*argv+2);
		else if (strncmp(*argv, "-f",2)==0) freq= atoi(*argv+2);
		else if (strcmp(*argv, "-b")==0) be=  1;
		else if (strcmp(*argv, "-m")==0) noc= 1;
		else if (strncmp(*argv,"-p",2)==0)ph= atoi(*argv+2);
		else if (strncmp(*argv,"-a",2)==0)alg= atoi(*argv+2);
		else if (strncmp(*argv,"-g",2)==0)gain= atoi(*argv+2);
		else if (strncmp(*argv,"-e",2)==0)echo= atoi(*argv+2);
        	else if (strcmp(*argv, "-") ==0) bs= stdin;
        	else if (**argv == '-') {
                        fprintf(stderr, "unknown option: %s\n", *argv);
                        usage(prg);
                        }
                else break;
                argc--; argv++;
                }

	mixer(volume);

       	if ((fp= fopen(IR,"r")) == NULL){
		fprintf(stderr, "can't open %s\n",IR);
		exit(1);
		}
	irlen= fread(spl,1,sizeof(spl),fp)/sizeof(short);
	fclose(fp);
	norm_ir();

	if(alg==7) polinv();
	if(alg==3){ acor(spl,irlen); lpc(irlen); }
	acor(spl,irlen); lpc(irlen);

	if(argc==0){ if(bs!=stdin)usage(prg); aout(); }
	else while(argc--){
       		if ((bs= fopen(*argv++,"r"))==NULL)
			{ fprintf(stderr,"%s open err",*(--argv)); exit(1); }
		aout();
		}
	}
/*
** audio_fwrite
*/
audio_fwrite(fd,bf,l)
char *bf;
	{
	int n;
	while(l) {
		if((n= write (fd,bf,l))<0){
			if(errno != EINTR)
			{ perror("write to /dev/dsp failed"); exit(1); }
			n= 0;
			}
		bf+= n; l-= n;
		}
	}
/*
** signal expand
*/
expand(int fd,short *sp) {
	int i;
	for(i=0;i<bz/2/rvl;i++) {
		audio_fwrite(fd,sp+i*rvl,rvl*4);
		audio_fwrite(fd,sp+i*rvl,rvl*4);
		}
	}

/*
** read_ac()
*/
read_ac() {
	FILE *fp;
	if (fp= fopen(AC,"r")) {
		fscanf(fp,"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
		&alg,&gain,&ph,&rv,&rvl,&echo,&echolen,&coeff,&diff,&expd, 
		&ffir,fa,fa+1,fa+2,fa+3,fa+4,fa+5,fa+6,fa+7,fa+8,fa+9);
		fclose(fp);
		}
	}
/*
** echop()
*/
echop(short *sp) {
	int i,of,t,l=echolen*441/10;
	l&= -2;
	for(i=0;i<bz;i++){
        	of= i-l;
		if(sp+of<ibf) {
			if(tail+of>=ibf) t= (tail[of]*coeff)>>7;
			}
		  else	 t= (sp[of]*coeff)>>7;
		t+= sp[i];
		if(t>MX)t=MX; if(t<MN)t=MN;
		sp[i]= t;
		}
	}

/*
** rev()
*/
rev(int dif,short *sp) {
	int i,j,h;
	int   *ip= (int *)sp;
	static short last0= 0;
	static short last1= 0;
	short newlast0;
	short newlast1;
	int t;

	if(dif) {
		newlast0= sp[bz-2];
		newlast1= sp[bz-1];
		for(i=bz-1;i>=2;i--)
		{ t= sp[i]-sp[i-2]; if(t>MX)t=MX; if(t<MN)t=MN; sp[i]= t; }
		t= sp[0]-last0; if(t>MX)t=MX; if(t<MN)t=MN; sp[0]= t;
		t= sp[1]-last1; if(t>MX)t=MX; if(t<MN)t=MN; sp[1]= t;
		}
	for(i=0;i<bz/2/rvl;i++) {
		int *ip0= ip+i*rvl;
		int *ip1= ip0+rvl-1;
		for(j=0;j<rvl/2;j++) { int t= ip0[j]; ip0[j]= ip1[-j]; ip1[-j]= t; }
		}
	if(dif){
		t= sp[0]+last0; if(t>MX)t=MX; if(t<MN)t=MN; sp[0]= t;
		t= sp[1]+last1; if(t>MX)t=MX; if(t<MN)t=MN; sp[1]= t;
		for(i=2;i<bz;i++)
		{ t= sp[i]+sp[i-2]; if(t>MX)t=MX; if(t<MN)t=MN; sp[i]= t; }
		last0= newlast0;
		last1= newlast1;
		}
	}

/*
**	phaser()
*/
phaser(short *sp) {
	int i,j,k,l,r,t;
	for(i=0;i<bz;i+=2){
		l= sp[i]; r= sp[i+1];
		t= l - ((r*ph)>>4);
		if(t>MX)t= MX; if(t<MN)t= MN; sp[i]= t;
		t= r - ((l*ph)>>4);
		if(t>MX)t= MX; if(t<MN)t= MN; sp[i+1]= t;
		}
	}

/*
** audio out
*/
aout() {
	char *bf;
	int i,l,fd,lbz;
	short *sp = ibf;

	fd= open_audio(noc,freq);
	fread((char *)ibf,1,48,bs);

	bz= (BS/2)/rvl*rvl*2;
	while(l= fread(sp, 4, bz/2, bs)*4){

		if(ph)phaser(sp);

		if(rv)rev(diff, sp);

		if(echo && echolen)echop(sp);

		for(i=0;i<bz;i++)buf[i+IS]= sp[i];

		if(alg==1) fir();
		if(alg==2) iir();
		if(alg==3||alg==7) ifir();

		if(ffir)fastFIR();

		if(expd)expand(fd,sp);
		else
		if(alg||ffir) audio_fwrite(fd,obf,l);
		  else  audio_fwrite(fd,sp,l);

		read_ac();
		bz= (BS/2)/rvl*rvl*2;
		if(sp+(l/2)+bz >= ibf+4*BS){ tail= sp+l/2; sp= ibf; ib= 0; }
		  else { sp+= bz; ib++; }
		}
	close(fd);
	}

/*
**	norm_ir()
**	normalize impulse respose 
**	method: average amplitude= 1<<10;
*/
norm_ir()
	{
	int i;
	double t;
	/*
	energy=0; for(i=0;i<irlen;i++){ int a= spl[i]; energy+= a*a; }
	t= sqrt(energy)/16384.;
	*/
	/*
	asum=0; for(i=0;i<irlen;i++){ int a= spl[i]; asum+= abs(a); }
	t= asum/65536.;
	for(i=0;i<irlen;i++){ int a= spl[i]; ipl[i]= a/t; }
	*/
	}

double c[MAXL]= {(1<<24)}; /* for ir3 and ir2 */

polinv()
	{
	int i,j,k;
	double r;
	for(j=0;j<irlen;j++){
		ipl[j]= r= c[j]/spl[0];
		if(abs(ipl[j])>100000) break;
		fprintf(stderr, "ipl[%d]=%d\n",j,ipl[j]);
		for(i=j;i<irlen;i++) c[i]-= r * spl[i-j];
		}
	for(;j<irlen;j++)ipl[j]= 0;
	}


fastFIR() {
	short i,j,k;
	for(j=IS;j<IS+bz;j+=2){
		int s0= 0;
		int s1= 0;
		for(k=j-10*2-2,i=9;i>=0;i--,k+=2){
			s0+=fa[i]*buf[k  ];
			s1+=fa[i]*buf[k+1];
			}
		s0>>= 7;
		s1>>= 7;
		if(s0>MX)s0= MX; if(s0<MN)s0= MN;
		if(s1>MX)s1= MX; if(s1<MN)s1= MN;
		obf[j-IS]=   s0;
		obf[j+1-IS]= s1;
		}
	memcpy (buf, buf+bz, IS*sizeof(int));
	}

fir()
{
	short i,j,k;
	for(j=IS;j<IS+BS;j+=2){
		int s0= 0;
		int s1= 0;
		for(k=j-irlen*2-2,i=irlen-1;i>=0;i--,k+=2){
			s0+=spl[i]*buf[k  ];
			s1+=spl[i]*buf[k+1];
		}
		s0>>= gain;
		s1>>= gain;
		if(s0>MX)s0= MX; if(s0<MN)s0= MN;
		if(s1>MX)s1= MX; if(s1<MN)s1= MN;
		obf[j-IS]=   s0;
		obf[j+1-IS]= s1;
	}
	memcpy (buf, buf+BS, IS*sizeof(int));
}

ifir()
{
	short i,j,k;
	for(j=IS;j<IS+BS;j+=2){
		int s0= 0;
		int s1= 0;
		for(k=j-irlen*2-2,i=irlen-1;i>=0;i--,k+=2){
			s0+=ipl[i]*buf[k  ];
			s1+=ipl[i]*buf[k+1];
		}
		s0>>= gain;
		s1>>= gain;
		if(s0>MX)s0= MX; if(s0<MN)s0= MN;
		if(s1>MX)s1= MX; if(s1<MN)s1= MN;
		obf[j-IS]= s0;
		obf[j+1-IS]= s1;
	}
	memcpy (buf, buf+BS, IS*sizeof(int));
}

/*
** iir()
*/
iir()
{
	short i,k,j;
	static int a0= 0;
	if(a0==0){ a0= ipl[0]; ipl[0]= 0; }

	for(j=IS;j<IS+BS;j+=2){
		int s0= 0;
		int s1= 0;
		for(k=j-irlen*2-2,i=irlen-1;i>0;i--,k+=2){
			s0+=ipl[i]*buf[k  ];
			s1+=ipl[i]*buf[k+1];
		}
		s0>>= gain; s1>>= gain;
		if(a0<0){ s0= buf[j]+s0; s1= buf[j+1]+s1; }
		   else { s0= buf[j]-s0; s1= buf[j+1]-s1; }
		if(s0>MX)s0= MX; if(s0<MN)s0= MN;
		if(s1>MX)s1= MX; if(s1<MN)s1= MN;
		obf[j-IS]= buf[j   ]= s0;
		obf[j+1-IS]= buf[j+1   ]= s1;
	}
	memcpy (buf, buf+BS, IS*sizeof(int));
}

big_endian() { int i; unsigned int d;
for(i=0;i<BS;i++){ d= (unsigned short)buf[i]; buf[i]= d<<8 | d>>8; }
}

open_audio(noc,freq)
int noc;	/* nomber_of_channels      */
int freq;	/* sampling frequency [kHz]*/
{
int i, fd, flags, ret, vol_chan;

static int play_precision;
static int play_stereo;
static int play_sample_rate;


  if ((fd= open ("/dev/dsp", O_WRONLY | O_NDELAY, 0)) < 0){
      fprintf(stderr,"audio device /dev/dsp is busy or can't be opened\n");
      exit(1);
    }

  /* turn NDELAY mode off: */
  if ((flags = fcntl (fd, F_GETFL, 0)) < 0) {
    fprintf(stderr,"fcntl F_GETFL on /dev/audio failed\n");
    exit(1);
  }
  flags &= ~O_NDELAY;
  if (fcntl (fd, F_SETFL, flags) < 0) {
    fprintf(stderr,"fcntl F_SETFL on /dev/audio failed\n");
    exit(1);
  }

  /* configure the device: */
  play_precision= 16;
  play_stereo= noc-1;
  play_sample_rate= freq;
  if(
      ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &play_precision) == -1 ||
      ioctl(fd, SNDCTL_DSP_STEREO, &play_stereo) == -1 ||
      ioctl(fd, SNDCTL_DSP_SPEED, &play_sample_rate) == -1
    )
  { fprintf(stderr,"configuration of /dev/dsp failed\n"); exit (1); }
return fd;
} 

mixer(volume)
int volume;/* volume 0-100 */
{
  int fd, ret, devices, vol_chan;
  if ((fd = open ("/dev/mixer", O_WRONLY | O_NDELAY, 0)) < 0){
      fprintf(stderr,"audio device /dev/mixer is busy or can't be opened\n");
      exit(1);
    }
  ret = ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devices);
  if (ret == -1) devices = 0;
  if (devices & (1 << SOUND_MIXER_PCM))
              vol_chan = SOUND_MIXER_PCM;
  else        vol_chan = SOUND_MIXER_VOLUME;
  volume |= volume << 8;
  if( ioctl(fd, MIXER_WRITE(vol_chan), &volume) == -1)
  { fprintf(stderr,"configuration of /dev/mixer failed\n"); exit (1); }
  close(fd);
}

double ac[MAXL];
double al[MAXL];
double kp[MAXL];

/*
**	acor()
**
**	autocorrelation
**	input signal is short (signed 16 bit)
**	analizing length is n.
*/
acor(sig, n)
short *sig;
	{
	int i, j;
	for(j=0;j<n;j++){
		double sum= 0;
		for(i=0;i<n-j;i++) sum+= sig[i]*sig[i+j];
		ac[j]= sum/n;
		}
	}

/*
**	lpc()
**	calculate alpha parameters and k parameters
**	<<< Levinson and Durbin method >>>
**	input:	double ac[N];
**	output:	double al[N];
**		double kp[N];
*/
lpc(p)
	{
	int i,j;
	double u,w,k;

	al[0]= 1.0;
	al[1]= 0.0;

	w=  ac[1];	/* w= ? */ /* w0= ac1 */
	u=  ac[0];	/* u= E */ /* u0= ac0 */

	for(j= 1; j <= p; j++){

		k= kp[j]= w/u;

		/*
		k1= ac1/ac0,
		k2= w1/u1= (ac2-ac1^2/ac0)/(ac0-ac1^2/ac0) = (ac2*ac0-ac1^2)/(ac0^2-ac1^2)
		*/

		u-= k*w;	/* E= (1-k^2)E */

		/* u1= u0-k1*ac1= ac0-ac1^2/ac0 */
		/* u2= u1-k2*w1=  u1 - w1^2/u1 */

		/*
		for(i= 1;i <= j; i++){
			al[i]-= k*al[j-i]; Note using old value
			}
		*/
	/* j= 1: al[1]-= k1*al[0]==> al[1]= -k1		*/
	/* j= 2: al[1]-= k2*al[1], al[2]-= k2*al[0]		*/
	/* j= 3: al[1]-= k3*al[2], al[2]-= k3*al[1], al[3]-=k3*al[0]	*/
	/* j= 4: al[1]-= k4*al[3], al[2]-= k4*al[2], al[3]-=k4*al[1], al[4]-=k4*al[0] */

		al[j]= -k;
		if(j>1){
			if(j>2){
				for(i=1;i<(j+1)/2;i++){
					double t1= al[i  ];
					double t2= al[j-i];
					al[i  ]= t1 - k*t2;
					al[j-i]= t2 - k*t1;
					}
				}
			if((j&1)==0) al[j/2]*= 1. - k;
			}
		w= 0.0;
		for(i=0;i<=j;i++) w+= al[i]*ac[j+1-i];
		/* w1= ac2-k1*ac1 = ac2- ac1^2/ac0 */
		/* w2= al[0]*ac[3] +al[1]*ac[2] +al[2]*ac[1] */
		}

	for(i=0;i<p;i++) {
		ipl[i]= -al[i]*(1<<14);
		/*
		printf("%d : al= %f  %d kp= %f\n",i, al[i], ipl[i], kp[i]);
		*/
		}
	/*
	ipl[0]= (1<<14);
	ipl[0]= 0;
	gain= 15;
	*/
	gain= 10;
	}

